# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from karabogui.widgets.hints import ElidingLabel


def test_elided_label(gui_app):
    label = ElidingLabel()
    label.setMaximumWidth(100)
    label.setText("Karabo")
    assert not label.isElided()
    label.setText("".join(15 * ["Karabo"]))
    assert label.isElided()
    label.setText("ThisIsKarabo")
    assert not label.isElided()
